#include <cstddef>
#include <cstdlib>
#include <iostream>
#include <vector>
#include <list>
#include <bitset>
#include <unistd.h>
#include <iomanip>
#include <sstream>
#include <cmath>
#include <functional>
#include <algorithm>
#include <err.h>

using namespace std;

bool flag_quiet = false;
bool flag_dont_print_ones = false;
bool flag_output_pbm = false;
bool flag_debug_recursion = false;
bool flag_debug_recursion_results = false;

const size_t max_value = 2000; 
typedef vector<size_t> num_vector;
typedef vector<num_vector> num_matrix;

typedef bitset<max_value> larger_than_N_bitset ;
typedef vector < larger_than_N_bitset > larger_than_N_matrix;

num_matrix gcds;
larger_than_N_matrix matrix;
num_matrix usable_pairs;

template<class COLLECTION>
inline
std::string
join_collection(const COLLECTION &col, const std::string& delimiter="\t")
{
	std::ostringstream os;
	bool first = true ;

	typename COLLECTION::const_iterator it = col.begin();

	while ( it != col.end() ) {
		if ( first ) 
			first = false;
		else
			os << delimiter ;
		os << *it ;
		it++ ;
	}

	return os.str();
}

size_t gcd ( size_t a, size_t b )
{
	size_t t;

	while ( b != 0 ) {
		t = b;
		b = a % b ;
		a = t ;
	}
	return a;
}

void fill_gcd_vector ( const size_t A, const size_t B, num_vector& /*OUTPUT*/ gcd_vector )
{
	gcd_vector.clear();
	gcd_vector.resize ( B+1 ) ;
	gcd_vector[0] = 0 ;

	for ( size_t i = 1 ; i<=B ; ++i )
		gcd_vector[i] = gcd ( A, i ) ;
}

void fill_gcd_matrix ( const size_t A, num_matrix& /*OUTPUT*/ gcd_matrix )
{
	gcd_matrix.clear();
	gcd_matrix.resize ( A + 1 ) ;
	
	for ( size_t i = 1 ; i<=A; ++i )
		fill_gcd_vector ( i, A, gcd_matrix[i] ) ;
}

void fill_larger_than_N_vector ( const size_t index, const size_t T, const size_t N, larger_than_N_bitset& /*OUTPUT*/ bs )
{
	bs.reset();

	for ( size_t i = 1 ; i <= T ; ++i ) 
		bs.set( i, ( (i * index) / gcds[i][index] ) > N ) ;
}

void fill_larger_than_N_matrix ( const size_t T, const size_t N, larger_than_N_matrix& /*OUTPUT*/ matrix)
{
	matrix.resize(T+1);

	for ( size_t i = 1 ; i <= T; ++i)
		fill_larger_than_N_vector ( i, T, N, matrix[i] ) ;
}

/*
 * NOTE:
 *    indices are returned in decending order
 */
void get_indices ( const larger_than_N_bitset& bs, num_vector& /*OUTPUT*/ result )
{
	result.clear();
	for ( size_t i = bs.size()-1; i>=1 ; --i )
		if ( bs.test(i) )
			result.push_back(i);
}

void fill_usable_pairs ( const larger_than_N_matrix& matrix, num_matrix& /*OUTPUT*/ result )
{
	result.clear();

	result.resize ( matrix.size() ) ;

	for ( size_t i = 0 ; i < matrix.size(); ++i ) 
		get_indices ( matrix[i], result[i] );
}


list<size_t> current_numbers;
size_t	     recursion_level ;

list<size_t> max_product_numbers;
long long max_product ;


struct multiplier: public unary_function<size_t, void>
{
	multiplier() : product(1) {}
	long long product;
	void operator()(const size_t x) { product *= x; }
};

void find_other_numbers_recursive ( const num_vector& allowed_values, const size_t W )
{
	if ( allowed_values.empty() ) {
		if (flag_debug_recursion)
			cout << setw(recursion_level*4) << "" 
			     << "stopped at level " << recursion_level << endl;
		return ;
	}

	if ( W == 0 ) {
		if (flag_debug_recursion_results)
			cout << setw(recursion_level*4) << "" 
			     << "done: \t" << join_collection ( current_numbers, "," ) << endl;

		multiplier result = for_each ( current_numbers.begin(), current_numbers.end(),
						multiplier() ) ;

		if ( result.product > max_product ) {
			max_product_numbers = current_numbers ;
			max_product = result.product;

			cout  << "Product: \t" << max_product
			      << "\tnumbers:\t" << join_collection ( current_numbers, "," ) << endl;
		}
		return ;
	}

	for ( size_t i = 0; i < allowed_values.size(); ++i ) {
		size_t current_value = allowed_values[i];

		if (flag_debug_recursion)
			cout << setw(recursion_level*4) << "" << current_value << "," ;

		//Calculate intersecting usable pairs
		num_vector combined ( allowed_values.size() ) ;
		combined.erase(set_intersection ( allowed_values.begin(), allowed_values.end(),
				   usable_pairs[current_value].begin(),
				   usable_pairs[current_value].end(),
				   combined.begin(),
				   greater<size_t>()) ,
			       combined.end() ) ;

		if (flag_debug_recursion)
			cout << setw(recursion_level*4) << "" 
		   	     << "combined-list:\t" << join_collection ( combined, "," )
			     << endl;

		//Enter recursion level
		current_numbers.push_back( current_value ) ;
		recursion_level++;

		find_other_numbers_recursive ( combined, W - 1 ) ;

		//Exit recursion level
		recursion_level--;
		current_numbers.pop_back();
	}

	//if (recursion_level==3) 
	//	cerr << "." << endl ;
}

void find_other_numbers ( const size_t W )
{
	recursion_level = 1 ;
	max_product = 0 ;

	for ( size_t i = usable_pairs.size()-1; i>=1 ; --i ) {

		if ( usable_pairs[i].size() == 0 )
			continue ;

		if ( flag_debug_recursion )
			cout << i << endl;
		
		//Add this value
		current_numbers.push_back( i ) ;
		recursion_level++;

		find_other_numbers_recursive ( usable_pairs[i], W - 1 ) ;

		//Remove this value
		recursion_level--;
		current_numbers.pop_back();
	}
}

void usage()
{
	cout <<
"usage: find_other [-h] [-4] [-3] [-2] [-1] [-q] [-g A,B] [-a A,B] [-m A] [-o W,T,N] [-b I,T,N] [-k T,N] [-u T,N]\n" \
"by A. Gordon (gordon@cshl.edu)\n" \
"\n"\
"   -h     = This helpful help screen.\n" \
"   -q     = Quiet mode. No output is printed.\n" \
"            Useful for scripting. See -g for possible exit codes.\n" \
"   -1     = With -a,-m print space instead of 1 (for easier viewing).\n"  \
"   -2     = With -k, output PBM file (instead of textual matrix)\n" \
"   -3     = with -o, output recursion debug information\n" \
"   -4     = with -o, output recursion results\n" \
"   -g A,B = show GCD for A & B.\n" \
"            Exit with code  0 (SUCCESS) if GCD is 1.\n" \
"            Exit with code -1 (FAILURE) if GCD is larger than 1.\n" \
"            (Make sure there's no space between the values.)\n" \
"   -a A,B = list GCDs for A and (1..B)\n" \
"   -m A   = show GCD matrix of AxA integers \n" \
"   -o W,T,N = do that other thing...\n" \
"   -b I,T,N = print all Xs where:\n" \
"              1<X<=T   and\n" \
"              ((X*I)/gcd(X,I))>N\n" \
"   -k T,N   = print binary matrix, size TxT, where:\n" \
"              Mij = 1 iff\n" \
"              ((i*j)/gcd(i,j))>N\n" \
"   -u T,N   = for each X (1<X<=T), print usable pairs\n" \
"              based on the matrix from the '-k' option.\n" \
"\n" \
"Scripting examples:\n" \
"  $ find_other -q -g 5,11 && echo '5 and 11 are co-prime.'\n" \
"\n" ;

	exit(EXIT_FAILURE);
}

vector<size_t> parse_values ( const char* optarg )
{
	vector<size_t> results;
	size_t N;
	const char* input=optarg;
	char* endptr=NULL;

	while ( *input != 0 ) {
		N = strtoul ( input, &endptr, 10 ) ;

		if ( (*endptr!=0) && ( N==0 || *endptr != ',') )
			errx(1,"Invalid value specified (%s) format should be 'N1,N2...'", optarg);
//		if (N>max_value)
//			errx(1,"value (%zu) too large (max. value=%zu)", N, max_value);

		results.push_back(N);

		if ( *endptr == ',' )
			endptr++;
		
		input = endptr;
	}
	return results;
}

void do_show_gcd ( const char* optarg ) 
{
	vector<size_t> values = parse_values ( optarg );
	if ( values.empty() )
		errx(1,"Missing values on command line!");
	if ( values.size() != 2 )
		errx(1,"[-g] requires exactly two values on command line!");

	size_t result = gcd ( values[0], values[1] ) ;

	if (!flag_quiet)
		cout << result << endl;

	exit( (result==1) ? EXIT_SUCCESS : EXIT_FAILURE );
}

void do_show_all_gcds ( const char* optarg )
{
	vector<size_t> values = parse_values ( optarg );
	if ( values.empty() )
		errx(1,"Missing values on command line!");
	if ( values.size() != 2 )
		errx(1,"[-a] requires exactly two values on command line!");

	num_vector gcds;
	fill_gcd_vector ( values[0], values[1], gcds ) ;
	
	for ( size_t i = 1; i<=values[1]; ++i)
		if ( !flag_dont_print_ones || gcds[i]!=1 )
			cout << "gcd(" << values[0] << "," << i << ")\t" << gcds[i] << endl;

	exit(EXIT_SUCCESS);
}

void do_show_gcd_matrix ( const char* optarg )
{
	vector<size_t> values = parse_values ( optarg );
	if ( values.empty() )
		errx(1,"Missing values on command line!");
	if ( values.size() != 1 )
		errx(1,"[-m] requires exactly one value on command line!");

	size_t A = values[0];
	fill_gcd_matrix ( A, gcds ) ;
	
	cout << setw(6) << "" ;
	for ( size_t x = 1; x<=A; ++x)
		cout << setw(6) << x ;
	cout << endl;

	for ( size_t y = 1; y<=A; ++y) {
		cout << setw(6) << right << y ;
		for ( size_t x =1 ; x<=A; ++x)  {
			if ( !flag_dont_print_ones || gcds[y][x]!=1 )
				cout << setw(6) << gcds[y][x];
			else
				cout << setw(6) << "" ;
		}
		cout << endl;
	}

	exit(EXIT_SUCCESS);
}

void do_show_other( const char* optarg )
{
	vector<size_t> values = parse_values ( optarg );
	if ( values.empty() )
		errx(1,"Missing values on command line!");
	if ( values.size() != 3 )
		errx(1,"[-o] requires exactly three values on command line!");

	size_t W = values[0];
	size_t T = values[1];
	size_t N = values[2];

	fill_gcd_matrix ( T, gcds ) ;
	fill_larger_than_N_matrix ( T, N, matrix ) ;
	fill_usable_pairs ( matrix, usable_pairs ) ;

	find_other_numbers ( W ) ;

	exit(EXIT_SUCCESS);
}

void do_show_larger_than_N_bitset ( const char* optarg )
{
	vector<size_t> values = parse_values ( optarg );
	if ( values.empty() )
		errx(1,"Missing values on command line!");
	if ( values.size() != 3 )
		errx(1,"[-b] requires exactly three values on command line!");

	size_t Index = values[0];
	size_t T = values[1];
	size_t N = values[2];

	larger_than_N_bitset bs ;
	fill_gcd_matrix ( T, gcds ) ;
	fill_larger_than_N_vector ( Index,T, N, bs ) ;

	num_vector indices;
	get_indices ( bs, indices ) ;

	cout << join_collection ( indices, ", " ) << endl;

	exit(EXIT_SUCCESS);
}

void do_show_larger_than_N_matrix ( const char* optarg )
{
	vector<size_t> values = parse_values ( optarg );
	if ( values.empty() )
		errx(1,"Missing values on command line!");
	if ( values.size() != 2 )
		errx(1,"[-k] requires exactly two values on command line!");

	size_t T = values[0];
	size_t N = values[1];

	fill_gcd_matrix ( T, gcds ) ;

	fill_larger_than_N_matrix ( T, N, matrix ) ;

	/* Display the matrix */
	if ( flag_output_pbm ) {
		cout << "P1" << endl;
		cout << "#matrix.pbm" << endl;
		cout << T << " " << T << endl;

		size_t pos_in_line = 0 ;
		for ( size_t i = 1 ; i <= T ; ++i ) {
			for ( size_t b = 1 ; b <= T; ++b )  {
				cout << (matrix[i].test(b)?"1":"0") << " ";
				pos_in_line++;
				if (pos_in_line>68){
					cout << endl;
					pos_in_line=0 ;
				}
			}
		}
	}
	else {
		cout << "-\t"; 
		for ( size_t i = 0 ; i <= T ; i++ ) {
			if ( i%10==0 )
				cout << (i/10) ;
			else
				cout << " " ;
		}
		cout << endl;
		for ( size_t i = 0 ; i <= T ; ++i ) {
			cout << i << "\t" ;
			for ( size_t b = 0 ; b <= T; ++b ) 
				cout << (matrix[i].test(b)?"1":" ") ;
			cout << endl;
		}
	}


	exit(EXIT_SUCCESS);
}

void do_show_usable_pairs ( const char* optarg ) 
{
	vector<size_t> values = parse_values ( optarg );
	if ( values.empty() )
		errx(1,"Missing values on command line!");
	if ( values.size() != 2 )
		errx(1,"[-u] requires exactly two values on command line!");

	size_t T = values[0];
	size_t N = values[1];

	fill_gcd_matrix ( T, gcds ) ;
	fill_larger_than_N_matrix ( T, N, matrix ) ;
	fill_usable_pairs ( matrix, usable_pairs ) ;

	for (size_t i=1;i<=T;++i) {
		if ( usable_pairs[i].size()==0 )
			continue;

		cout << setw(6) << i << ":\t"; 
		cout << join_collection ( usable_pairs[i], "," ) << endl;
	}

	exit(EXIT_SUCCESS);
}


int main(int argc, char* argv[])
{
	int opt;
	while ( (opt=getopt(argc,argv,"4321qhg:a:m:o:b:k:u:"))!=-1 ) {
		switch(opt)
		{
		case 'q':
			flag_quiet = true;
			break;

		case '1':
			flag_dont_print_ones = true;
			break;

		case '2':
			flag_output_pbm = true;
			break;

		case '3':
			flag_debug_recursion = true;
			break;

		case '4':
			flag_debug_recursion_results = true;
			break;

		case 'h':
			usage();

		case 'g':
			do_show_gcd ( optarg ) ;

		case 'a':
			do_show_all_gcds ( optarg ) ;

		case 'm':
			do_show_gcd_matrix ( optarg ) ;

		case 'o':
			do_show_other ( optarg ) ;

		case 'b':
			do_show_larger_than_N_bitset ( optarg ) ;

		case 'k':
			do_show_larger_than_N_matrix ( optarg ) ;

		case 'u':
			do_show_usable_pairs ( optarg ) ;
		}
	}

	usage();

	return 0;
}
